import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import warnings

from wildboottest.wildboottest import WildboottestCL, WildboottestHC
from importlib import import_module
from typing import Union, List, Dict
from scipy.stats import norm, t
from pyfixest.ssc_utils import get_ssc


class Feols:

    """
    A class for estimating regression models with high-dimensional fixed effects via
    ordinary least squares.

    Parameters
    ----------
    Y : Union[np.ndarray, pd.DataFrame]
        Dependent variable of the regression.
    X : Union[np.ndarray, pd.DataFrame]
        Independent variable of the regression.
    Z: Union[np.ndarray, pd.DataFrame]
        Instruments of the regression.

    Attributes
    ----------
    Y : np.ndarray
        The dependent variable of the regression.
    X : np.ndarray
        The independent variable of the regression.
    Z : np.ndarray
        The instruments of the regression.
    N : int
        The number of observations.
    k : int
        The number of columns in X.

    Methods
    -------
    get_fit()
        Regression estimation for a single model, via ordinary least squares (OLS).
    get_vcov(vcov)
        Compute covariance matrices for an estimated model.

    Raises
    ------
    AssertionError
        If the vcov argument is not a dict, a string, or a list.

    """

    def __init__(self, Y: np.ndarray, X: np.ndarray, Z: np.ndarray) -> None:

        if not isinstance(Y, (np.ndarray)):
            raise TypeError("Y must be a numpy array.")
        if not isinstance(X, (np.ndarray)):
            raise TypeError("X must be a numpy array.")

        self.Y = Y
        self.X = X
        self.Z = Z

        if self.X.ndim != 2:
            raise ValueError("X must be a 2D array")
        if self.Z.ndim != 2:
            raise ValueError("Z must be a 2D array")

        self.N, self.k = X.shape

    def get_fit(self) -> None:
        '''
        Regression estimation for a single model, via ordinary least squares (OLS).
        '''

        self.tZX = np.transpose(self.Z) @ self.X
        self.tZXinv = np.linalg.inv(self.tZX)

        self.tZy = (np.transpose(self.Z) @ self.Y)
        beta_hat = self.tZXinv @ self.tZy
        self.beta_hat = beta_hat.flatten()
        self.Y_hat = (self.X @ self.beta_hat)
        self.u_hat = (self.Y.flatten() - self.Y_hat)

    def get_vcov(self, vcov: Union[str, Dict[str, str], List[str]]) -> None:
        '''
        Compute covariance matrices for an estimated regression model.

        Parameters
        ----------
        vcov : Union[str, Dict[str, str], List[str]]
            A string or dictionary specifying the type of variance-covariance matrix to use for inference.
            If a string, can be one of "iid", "hetero", "HC1", "HC2", "HC3".
            If a dictionary, it should have the format {"CRV1":"clustervar"} for CRV1 inference
            or {"CRV3":"clustervar"} for CRV3 inference.
            Note that CRV3 inference is currently not supported with arbitrary fixed effects and IV estimation.

        Raises
        ------
        AssertionError
            If vcov is not a dict, string, or list.
        AssertionError
            If vcov is a dict and the key is not "CRV1" or "CRV3".
        AssertionError
            If vcov is a dict and the value is not a string.
        AssertionError
            If vcov is a dict and the value is not a column in the data.
        AssertionError
            CRV3 currently not supported with arbitrary fixed effects
        AssertionError
            If vcov is a list and it does not contain strings.
        AssertionError
            If vcov is a list and it does not contain columns in the data.
        AssertionError
            If vcov is a string and it is not one of "iid", "hetero", "HC1", "HC2", or "HC3".


        Returns
        -------
        None

        '''

        assert isinstance(vcov, (dict, str, list)), "vcov must be a dict, string or list"
        if isinstance(vcov, dict):
            assert list(vcov.keys())[0] in ["CRV1", "CRV3"], "vcov dict key must be CRV1 or CRV3"
            assert isinstance(list(vcov.values())[0], str), "vcov dict value must be a string"
            assert list(vcov.values())[0] in self.data.columns, "vcov dict value must be a column in the data"
        if isinstance(vcov, list):
            assert all(isinstance(v, str) for v in vcov), "vcov list must contain strings"
            assert all(v in self.data.columns for v in vcov), "vcov list must contain columns in the data"
        if isinstance(vcov, str):
            assert vcov in ["iid", "hetero", "HC1", "HC2", "HC3"], "vcov string must be iid, hetero, HC1, HC2, or HC3"

        if isinstance(vcov, dict):
            vcov_type_detail = list(vcov.keys())[0]
            self.clustervar = list(vcov.values())[0]
        elif isinstance(vcov, list):
            vcov_type_detail = vcov
        elif isinstance(vcov, str):
            vcov_type_detail = vcov
        else:
            assert False, "arg vcov needs to be a dict, string or list"

        if vcov_type_detail == "iid":
            self.vcov_type = "iid"
            self.is_clustered = False
        elif vcov_type_detail in ["hetero", "HC1", "HC2", "HC3"]:
            self.vcov_type = "hetero"
            self.is_clustered = False
            if vcov_type_detail in ["HC2", "HC3"]:
                if self.has_fixef:
                    raise ValueError("HC2 and HC3 inference types are not supported for regressions with fixed effects.")
        elif vcov_type_detail in ["CRV1", "CRV3"]:
            self.vcov_type = "CRV"
            self.is_clustered = True


        if self.is_iv:
            if self.vcov_type in ["CRV3"]:
                raise ValueError("CRV3 inference is not supported for IV regressions.")

        # compute vcov
        if self.vcov_type == 'iid':

            self.ssc = get_ssc(
                ssc_dict = self.ssc_dict,
                N = self.N,
                k = self.k,
                G = 1,
                vcov_sign = 1,
                vcov_type='iid'
            )

            # only relevant factor for iid in ssc: fixef.K
            if self.is_iv == False:
                self.vcov =  self.ssc * self.tZXinv * (np.sum(self.u_hat ** 2) / (self.N - 1))
            else:
                sigma2 = (np.sum(self.u_hat ** 2) / (self.N - 1))
                tZZinv = np.linalg.inv(np.transpose(self.Z) @ self.Z) # k x k
                tXZ = np.transpose(self.X) @ self.Z
                self.vcov = self.ssc * np.linalg.inv(tXZ @ tZZinv @ self.tZX ) * sigma2 #

        elif self.vcov_type == 'hetero':

            self.ssc = get_ssc(
                ssc_dict = self.ssc_dict,
                N = self.N,
                k = self.k,
                G = 1,
                vcov_sign = 1,
                vcov_type = "hetero"
            )

            if vcov_type_detail in ["hetero", "HC1"]:
                u = self.u_hat
            elif vcov_type_detail in ["HC2", "HC3"]:
                leverage = np.sum(self.X * (self.X @ self.tZXinv), axis=1)
                if vcov_type_detail == "HC2":
                    u = self.u_hat / np.sqrt(1 - leverage)
                else:
                    u = self.u_hat / (1-leverage)

            if self.is_iv == False:
                meat = np.transpose(self.Z) * (u ** 2) @ self.Z
                # set off diagonal elements to zero
                self.vcov =  self.ssc * self.tZXinv @ meat @  self.tZXinv
            else:
                tZZinv = np.linalg.inv(np.transpose(self.Z) @ self.Z)  # k x k
                tXZ = np.transpose(self.X) @ self.Z # k x k
                if u.ndim == 1:
                    u = u.reshape((self.N,1))
                Omega = np.transpose(self.Z) @ (self.Z * (u ** 2))  # k x k
                meat = tXZ @ tZZinv  @ Omega  @ tZZinv @ self.tZX # k x k
                bread = np.linalg.inv(tXZ @ tZZinv @ self.tZX)
                self.vcov = self.ssc * bread @ meat @ bread



        elif self.vcov_type == "CRV":

            cluster_df = self.data[self.clustervar]
            # if there are missings - delete them!

            if cluster_df.dtype != "category":
                cluster_df = pd.Categorical(cluster_df)

            if cluster_df.isna().any():
                raise ValueError("CRV inference not supported with missing values in the cluster variable. Please drop missing values before running the regression.")

            cluster_mat, clustid = pd.factorize(cluster_df)
            #cluster_mat = cluster_df.to_numpy()

            #clustid = np.unique(cluster_mat)
            self.G = len(clustid)

            self.ssc = get_ssc(
                ssc_dict = self.ssc_dict,
                N = self.N,
                k = self.k,
                G = self.G,
                vcov_sign = 1,
                vcov_type = "CRV"
            )

            if vcov_type_detail == "CRV1":


                meat = np.zeros((self.k, self.k))

                for _, g, in enumerate(clustid):

                    Zg = self.Z[np.where(cluster_df == g)]
                    ug = self.u_hat[np.where(cluster_df == g)]
                    score_g = (np.transpose(Zg) @ ug).reshape((self.k, 1))
                    meat += np.dot(score_g, score_g.transpose())

                if self.is_iv == False:
                    self.vcov = self.ssc * self.tZXinv @ meat @ self.tZXinv
                else:
                    tZZinv = np.linalg.inv(np.transpose(self.Z) @ self.Z)  # k x k
                    tXZ = np.transpose(self.X) @ self.Z # k x k
                    meat = tXZ @ tZZinv @ meat @ tZZinv @ self.tZX
                    bread = np.linalg.inv(tXZ @ tZZinv @ self.tZX)
                    self.vcov = self.ssc * bread @ meat @ bread

            elif vcov_type_detail == "CRV3":

                # check: is fixed effect cluster fixed effect?
                # if not, either error or turn fixefs into dummies
                # for now: don't allow for use with fixed effects

                #if self.has_fixef:
                #    raise ValueError("CRV3 inference is currently not supported with fixed effects.")

                if self.is_iv:
                    raise ValueError("CRV3 inference is not supported with IV estimation.")

                k_params = self.k

                beta_hat = self.beta_hat

                clusters = clustid
                n_groups = self.G
                group = cluster_df

                beta_jack = np.zeros((n_groups, k_params))


                if self.has_fixef == False:
                    # inverse hessian precomputed?
                    tXX = np.transpose(self.X) @ self.X
                    tXy = np.transpose(self.X) @ self.Y

                    # compute leave-one-out regression coefficients (aka clusterjacks')
                    for ixg, g in enumerate(clusters):

                        Xg = self.X[np.equal(ixg, group)]
                        Yg = self.Y[np.equal(ixg, group)]
                        tXgXg = np.transpose(Xg) @ Xg
                        # jackknife regression coefficient
                        beta_jack[ixg,:] = (
                            np.linalg.pinv(tXX - tXgXg) @ (tXy - np.transpose(Xg) @ Yg)
                        ).flatten()

                else:

                    # lazy loading to avoid circular import
                    fixest_module = import_module('pyfixest.fixest')
                    Fixest_ = getattr(fixest_module, 'Fixest')

                    for ixg, g in enumerate(clusters):
                        # direct leave one cluster out implementation
                        data = self.data[~np.equal(ixg, group)]
                        model = Fixest_(data)
                        model.feols(self.fml, vcov = "iid")
                        beta_jack[ixg,:] = model.coef()["Estimate"].to_numpy()


                # optional: beta_bar in MNW (2022)
                #center = "estimate"
                #if center == 'estimate':
                #    beta_center = beta_hat
                #else:
                #    beta_center = np.mean(beta_jack, axis = 0)
                beta_center = beta_hat

                vcov = np.zeros((k_params, k_params))
                for ixg, g in enumerate(clusters):
                    beta_centered = beta_jack[ixg,:] - beta_center
                    vcov += np.outer(beta_centered, beta_centered)

                self.vcov = self.ssc * vcov

    def get_inference(self, alpha = 0.95):
        '''
        Compute standard errors, t-statistics and p-values for the regression model.

        Parameters
        ----------
        alpha : float, optional
            Significance level for confidence intervals, by default 0.95

        Returns
        -------
        None

        '''

        self.se = (
            np.sqrt(np.diagonal(self.vcov))
        )

        self.tstat = (
            self.beta_hat / self.se
        )

        #if self.vcov_type in ['iid', 'CRV']:
            # t(G-1) distribution for clustered errors
        if self.vcov_type in ["iid", "hetero"]:
            df = self.N - self.k
        else:
            df = self.G - 1
        self.pvalue = (
            2*(1-t.cdf(np.abs(self.tstat), df))
        )
        #else:
        #    # normal distribution for non-clustered errors
        #    self.pvalue = (
        #            2*(1-norm.cdf(np.abs(self.tstat)))
        #    )

        z = norm.ppf(1 - (alpha / 2))
        self.conf_int = (
            np.array([z * self.se - self.beta_hat, z * self.se + self.beta_hat])
        )


    def get_wildboottest(self, B:int, cluster : Union[np.ndarray, pd.Series, pd.DataFrame, None], param : Union[str, None], weights_type: str, impose_null: bool , bootstrap_type: str, seed: Union[str, None] , adj: bool , cluster_adj: bool):

        '''
        Run a wild cluster bootstrap based on an object of type "Feols"

        Args:

        B (int): The number of bootstrap iterations to run
        cluster (Union[None, np.ndarray, pd.Series, pd.DataFrame], optional): If None (default), a 'heteroskedastic' wild boostrap
            is run. For a wild cluster bootstrap, requires a numpy array of dimension one,a  pandas Series or DataFrame, containing the clustering variable.
        param (Union[str, None], optional): A string of length one, containing the test parameter of interest. Defaults to None.
        weights_type (str, optional): The type of bootstrap weights. Either 'rademacher', 'mammen', 'webb' or 'normal'.
                            'rademacher' by default. Defaults to 'rademacher'.
        impose_null (bool, optional): Should the null hypothesis be imposed on the bootstrap dgp, or not?
                            Defaults to True.
        bootstrap_type (str, optional):A string of length one. Allows to choose the bootstrap type
                            to be run. Either '11', '31', '13' or '33'. '11' by default. Defaults to '11'.
        seed (Union[str, None], optional): Option to provide a random seed. Defaults to None.

        Returns: a pd.DataFrame with bootstrapped t-statistic and p-value
        '''

        if self.is_iv:
            raise ValueError("Wild cluster bootstrap is not supported with IV estimation.")
        if self.has_fixef:
            raise ValueError("Wild cluster bootstrap is not supported with fixed effects.")

        xnames = self.coefnames.to_list()
        Y = self.Y.flatten()
        X = self.X

        # later: allow r <> 0 and custom R
        R = np.zeros(len(xnames))
        R[xnames.index(param)] = 1
        r = 0

        if cluster is None:

            boot = WildboottestHC(X = X, Y = Y, R = R, r = r, B = B, seed = seed)
            boot.get_adjustments(bootstrap_type = bootstrap_type)
            boot.get_uhat(impose_null = impose_null)
            boot.get_tboot(weights_type = weights_type)
            boot.get_tstat()
            boot.get_pvalue(pval_type = "two-tailed")
            full_enumeration_warn = False

        else:

            cluster = self.data[self.clustervar]

            boot = WildboottestCL(X = X, Y = Y, cluster = cluster,
                                R = R, B = B, seed = seed)
            boot.get_scores(bootstrap_type = bootstrap_type, impose_null = impose_null, adj=adj, cluster_adj=cluster_adj)
            _, _, full_enumeration_warn = boot.get_weights(weights_type = weights_type)
            boot.get_numer()
            boot.get_denom()
            boot.get_tboot()
            boot.get_vcov()
            boot.get_tstat()
            boot.get_pvalue(pval_type = "two-tailed")

            if full_enumeration_warn:
                warnings.warn("2^G < the number of boot iterations, setting full_enumeration to True.")

        res = {
            'param':param,
            'statistic': boot.t_stat,
            'pvalue': boot.pvalue,
            'bootstrap_type': bootstrap_type,
            'impose_null' : impose_null
        }

        res_df = pd.Series(res)

        return res_df




    def get_nobs(self):

        '''
        Fetch the number of observations used in fitting the regression model.

        Returns
        -------
        None
        '''
        self.N = len(self.Y)


    def get_performance(self):
        '''
        Compute multiple additional measures commonly reported with linear regression output.
        '''

        self.r_squared = 1 - np.sum(self.u_hat ** 2) / \
            np.sum((self.Y - np.mean(self.Y))**2)
        self.adj_r_squared = (self.N - 1) / (self.N - self.k) * self.r_squared



import pyhdfe
import re

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from typing import Any, Union, Dict, Optional, List, Tuple
from scipy.stats import norm
from formulaic import model_matrix

from pyfixest.feols import Feols
from pyfixest.FormulaParser import FixestFormulaParser, _flatten_list
from pyfixest.ssc_utils import ssc


class Fixest:

    def __init__(self, data: pd.DataFrame) -> None:
        '''
        A class for fixed effects regression modeling.
        Args:
            data: The input pd.DataFrame for the object.
        Returns:
            None
        '''

        self.data = data
        self.model_res = dict()

    def _clean_fe(self, data, fval):

        fval_list = fval.split("+")

        # find interacted fixed effects via "^"
        interacted_fes = [
            x for x in fval_list if len(x.split('^')) > 1]

        for x in interacted_fes:
            vars = x.split("^")
            data[x] = data[vars].apply(lambda x: '^'.join(
                x.dropna().astype(str)) if x.notna().all() else np.nan, axis=1)

        fe = data[fval_list]
        # all fes to factors / categories
        fe = fe.apply(lambda x: pd.factorize(x)[0])

        fe_na = fe.isna().any(axis=1)
        fe = fe.to_numpy()

        return fe, fe_na

    def _demean(self, data: pd.DataFrame, fval: str, ivars: List[str], drop_ref: str) -> None:
        '''
        Demean all regressions for a specification of fixed effects.

        Args:
            data: The input pd.DataFrame for the object. Either self.data or a subset thereof (for split sample estimation).
            fval: A specification of fixed effects. A string indicating the fixed effects to be demeaned,
                such as "X4" or "X3 + X2".
            ivars: A list of strings indicating the interacted variables via i().
            drop_ref: A string indicating the reference category for the interacted variables. The reference
                      category is dropped from the regression.

        Returns:
            None
        '''

        YXZ_dict = dict()
        na_dict = dict()
        var_dict = dict()

        if fval != "0":
            fe, fe_na = self._clean_fe(data, fval)
            fe_na = list(fe_na[fe_na == True])
        else:
            fe = None
            fe_na = None

        dict2fe = self.fml_dict2.get(fval)
        if self.is_iv:
            dict2fe_iv = self.fml_dict2_iv.get(fval)

        # create lookup table with NA index key
        # for each regression, check if lookup table already
        # populated with "demeaned" data for some (or all)
        # variables of the model. if demeaned variable for
        # NA key already exists -> use it. else rerun demeaning
        # algorithm

        lookup_demeaned_data = dict()

        # loop over both dict2fe and dict2fe_iv (if the latter is not None)
        for depvar in dict2fe.keys():

            # [(0, 'X1+X2'), (1, ['X1+X3'])]
            for _, covar in enumerate(dict2fe.get(depvar)):

                covar2 = covar
                depvar2 = depvar

                fml = depvar2 + " ~ " + covar2

                if self.is_iv:
                    instruments2 = dict2fe_iv.get(depvar)[0]
                    endogvar = list(set(covar2.split("+")) - set(instruments2.split("+")))[0]
                    instrument = list(set(instruments2.split("+")) - set(covar2.split("+")))[0]

                    if len([instrument]) > 1 or len([endogvar]) > 1:
                        raise ValueError("Currently, IV estimation is only supported with one endogeneous variable and one instrument.")

                    fml2 = instrument + "+" + fml
                    rhs, lhs = model_matrix(fml2, data)

                    Y = rhs[[depvar]]
                    I = rhs[[instrument]]
                    X = lhs
                else:
                    Y, X = model_matrix(fml, data)

                na_index = list(set(data.index) - set(Y.index))

                if self.ivars is not None:
                    if drop_ref is not None:
                        X = X.drop(drop_ref, axis=1)

                y_names = list(Y.columns)
                x_names = list(X.columns)
                yxz_names = list(y_names) + list(x_names)
                if self.is_iv:
                    iv_names = list(I.columns)
                    #z_varnames = list(set(x_names) - set([endogvar]))
                    #z_varnames.append(instrument)
                    x_names_copy = x_names.copy()
                    x_names_copy.remove(endogvar)
                    z_names = x_names_copy + [instrument]
                    cols = yxz_names + iv_names
                else:
                    iv_names = None
                    z_names = None
                    cols = yxz_names

                if self.ivars is not None:
                    self.icovars = [s for s in x_names if s.startswith(
                        ivars[0]) and s.endswith(ivars[1])]
                else:
                    self.icovars = None


                Y = Y.to_numpy()
                #if Y.ndim != 1:
                #    Y = Y.reshape((len(Y), 1))
                #.reshape((len(Y), 1))
                X = X.to_numpy()
                if self.is_iv:
                    I = I.to_numpy()
                    if I.ndim != 1:
                        I = I.reshape((len(I), 1))

                if Y.shape[1] > 1:
                    raise ValueError(
                        "Dependent variable must be a single column. Please make sure that the dependent variable" + depvar2 + "is of a numeric type (int or float).")

                # variant 1: if there are fixed effects to be projected out
                if fe is not None:
                    na_index = (na_index + fe_na)
                    fe2 = np.delete(fe, na_index, axis=0)
                    # drop intercept
                    intercept_index = x_names.index("Intercept")
                    X = np.delete(X, intercept_index, axis = 1)
                    #if self.is_iv:
                    #    I = np.delete(I, intercept_index, axis = 1)
                    x_names.remove("Intercept")
                    yxz_names.remove("Intercept")
                    if self.is_iv:
                        z_names.remove("Intercept")
                        cols.remove("Intercept")

                    # check if variables have already been demeaned
                    Y = np.delete(Y, fe_na, axis=0)
                    X = np.delete(X, fe_na, axis=0)
                    if self.is_iv:
                        I = np.delete(I, fe_na, axis=0)

                    if self.is_iv:
                        YXZ = np.concatenate([Y, X, I], axis = 1)
                    else:
                        YXZ = np.concatenate([Y, X], axis=1)

                    na_index_str = ','.join(str(x) for x in na_index)

                    # check if looked dict has data for na_index
                    if lookup_demeaned_data.get(na_index_str) is not None:
                        # get data out of lookup table: list of [algo, data]
                        algorithm, YXZ_demeaned_old = lookup_demeaned_data.get(
                            na_index_str)

                        # get not yet demeaned covariates
                        var_diff_names = list(
                            set(yxz_names) - set(YXZ_demeaned_old.columns))[0]
                        var_diff_index = list(yxz_names).index(var_diff_names)
                        var_diff = YXZ[:, var_diff_index]
                        if var_diff.ndim == 1:
                            var_diff = var_diff.reshape(len(var_diff), 1)

                        YXZ_demean_new = algorithm.residualize(var_diff)
                        YXZ_demeaned = np.concatenate(
                            [YXZ_demeaned_old, YXZ_demean_new], axis=1)
                        YXZ_demeaned = pd.DataFrame(YXZ_demeaned)

                        YXZ_demeaned.columns = list(
                            YXZ_demeaned_old.columns) + [var_diff_names]

                    else:
                        # not data demeaned yet for NA combination
                        algorithm = pyhdfe.create(
                            ids=fe2,
                            residualize_method='map',
                            drop_singletons=self.drop_singletons,
                        )
                        # n_singletons = algorithm.singletons
                        # if n_singletons is not None:
                        #    print(f'{n_singletons} columns are dropped due to singleton fixed effects.')
                        if self.drop_singletons == True and algorithm.singletons != 0 and algorithm.singletons is not None:
                            print(
                                algorithm.singletons, "columns are dropped due to singleton fixed effects.")
                            dropped_singleton_indices = (
                                np.where(algorithm._singleton_indices))[0].tolist()
                            na_index += dropped_singleton_indices

                        YXZ_demeaned = algorithm.residualize(YXZ)
                        YXZ_demeaned = pd.DataFrame(YXZ_demeaned)

                        #if self.is_iv:
                        #    cols += iv_names

                        YXZ_demeaned.columns = cols

                    lookup_demeaned_data[na_index_str] = [
                        algorithm, YXZ_demeaned]

                else:
                    # if no fixed effects
                    if self.is_iv:
                        YXZ = np.concatenate([Y, X, I], axis=1)
                    else:
                        YXZ = np.concatenate([Y, X], axis=1)

                    YXZ_demeaned = pd.DataFrame(YXZ)

                    #if self.is_iv:
                    #    cols += list(iv_names)

                    YXZ_demeaned.columns = cols

                YXZ_dict[fml] = YXZ_demeaned#[cols]
                na_dict[fml] = na_index
                var_dict[fml] = dict({
                    'y_names': y_names,
                    'x_names': x_names,
                    'iv_names': iv_names,
                    'z_names': z_names
                })


        return YXZ_dict, na_dict, var_dict

    def feols(self, fml: str, vcov: Union[None, str, Dict[str, str]] = None, ssc=ssc(), fixef_rm: str = "none") -> None:
        '''
        Method for fixed effects regression modeling using the PyHDFE package for projecting out fixed effects.
        Args:
            fml (str): A three-sided formula string using fixest formula syntax. Supported syntax includes:
                The syntax is as follows: "Y ~ X1 + X2 | FE1 + FE2 | X1 ~ Z1" where:

                Y: Dependent variable
                X1, X2: Independent variables
                FE1, FE2: Fixed effects
                Z1, Z2: Instruments
                |: Separates left-hand side, fixed effects, and instruments

                If no fixed effects and instruments are specified, the formula can be simplified to "Y ~ X1 + X2".
                If no instruments are specified, the formula can be simplified to "Y ~ X1 + X2 | FE1 + FE2".
                If no fixed effects are specified but instruments are specified, the formula can be simplified to "Y ~ X1 + X2 | X1 ~ Z1".

                Supported multiple estimation syntax includes:

                Stepwise regressions (sw, sw0)
                Cumulative stepwise regression (csw, csw0)
                Multiple dependent variables (Y1 + Y2 ~ X)

                Other special syntax includes:
                i() for interaction of a categorical and non-categorical variable (e.g. "i(X1,X2)" for interaction between X1 and X2).
                Using i() is required to use with some custom methods, e.g. iplot().
                ^ for interacted fixed effects (e.g. "fe1^fe2" for interaction between fe1 and fe2)

                All other parts of the formula must be compatible with formula parsing via the formulaic module.
                You can use formulaic functionaloty such as "C", "I", ":",, "*", "np.log", "np.power", etc.

            vcov (Union(str, dict)): A string or dictionary specifying the type of variance-covariance matrix to use for inference.
                If a string, it can be one of "iid", "hetero", "HC1", "HC2", "HC3".
                If a dictionary, it should have the format dict("CRV1":"clustervar") for CRV1 inference or dict(CRV3":"clustervar") for CRV3 inference.
            fixef_rm: A string specifiny whether singleton fixed effects should be dropped. Options are "none" (default) and "singleton". If "singleton", singleton fixed effects are dropped.
        Returns:
            None
        Examples:
            Standard formula:
                fml = 'Y ~ X1 + X2'
                fixest_model = Fixest(data=data).feols(fml, vcov='iid')
            With fixed effects:
                fml = 'Y ~ X1 + X2 | fe1 + fe2'
            With interacted fixed effects:
                fml = 'Y ~ X1 + X2 | fe1^fe2'
            Multiple dependent variables:
                fml = 'Y1 + Y2 ~ X1 + X2'
            Stepwise regressions (sw and sw0):
                fml = 'Y1 + Y2 ~ sw(X1, X2, X3)'
            Cumulative stepwise regressions (csw and csw0):
                fml = 'Y1 + Y2 ~ csw(X1, X2, X3) '
            Combinations:
                fml = 'Y1 + Y2 ~ csw(X1, X2, X3) | sw(X4, X5) + X6'
        '''

        self.fml = fml.replace(" ", "")
        self.split = None

        # deparse formula, at least partially
        fxst_fml = FixestFormulaParser(fml)

        if fxst_fml.is_iv:
            self.is_iv = True
        else:
            self.is_iv = False

        # add function argument to these methods for IV
        fxst_fml.get_fml_dict()
        fxst_fml.get_var_dict()
        fxst_fml._transform_fml_dict()

        if self.is_iv:
            # create required dicts for first stage IV regressions
            fxst_fml.get_fml_dict(iv = True)
            fxst_fml.get_var_dict(iv = True)
            fxst_fml._transform_fml_dict(iv = True)


        self.fml_dict = fxst_fml.fml_dict
        self.var_dict = fxst_fml.var_dict
        self.fml_dict2 = fxst_fml.fml_dict2

        if self.is_iv:
            self.fml_dict_iv = fxst_fml.fml_dict_iv
            self.var_dict_iv = fxst_fml.var_dict_iv
            self.fml_dict2_iv = fxst_fml.fml_dict2_iv
        else:
            self.fml_dict_iv = self.fml_dict
            self.var_dict_iv = self.var_dict
            self.fml_dict2_iv = self.fml_dict2

        self.ivars = fxst_fml.ivars


        self.ssc_dict = ssc
        if fixef_rm == "singleton":
            self.drop_singletons = True
        else:
            self.drop_singletons = False

        # get all fixed effects combinations
        fixef_keys = list(self.var_dict.keys())

        if self.ivars is not None:

            if list(self.ivars.keys())[0] is not None:
                ref = list(self.ivars.keys())[0]
                ivars = self.ivars[ref]
                drop_ref = ivars[0] + "[T." + ref + "]" + ":" + ivars[1]
            else:
                ivars = self.ivars[None]
                drop_ref = None

            # type checking
            i0_type = self.data[ivars[0]].dtype
            i1_type = self.data[ivars[1]].dtype
            if not i0_type in ['category', "O"]:
                raise ValueError("Column " + ivars[0] + " is not of type 'O' or 'category', which is required in the first position of i(). Instead it is of type " +
                                 i0_type.name + ". If a reference level is set, it is required that the variable in the first position of 'i()' is of type 'O' or 'category'.")
            if not i1_type in ['int64', 'float64', 'int32', 'float32']:
                raise ValueError("Column " + ivars[1] + " is not of type 'int' or 'float', which is required in the second position of i(). Instead it is of type " +
                                 i1_type.name + ". If a reference level is set, iti is required that the variable in the second position of 'i()' is of type 'int' or 'float'.")

        else:
            ivars = None
            drop_ref = None

        # dropped_data_dict and demeaned_data_dict are
        # dictionaries with keys for each fixed effects combination and
        # has values of lists of demeaned dataframes
        # the list is a singelton list unless split sample estimation is used
        # e.g it looks like this (without split estimation):
        # {'fe1': [demeaned_data_df], 'fe1+fe2': [demeaned_data_df]}
        # and like this (with split estimation):
        # {'fe1': [demeaned_data_df1, demeaned_data_df2], 'fe1+fe2': [demeaned_data_df1, demeaned_data_df2]}
        # the lists are sorted in the order of the split variable

        self.dropped_data_dict = dict()
        self.demeaned_data_dict = dict()
        # names of depvar, X, Z matrices
        self.yxz_name_dict = dict()

        estimate_full_model = True
        estimate_split_model = False
        # currently no fsplit allowed
        fsplit = None

        if self.split is not None:
            if fsplit is not None:
                raise ValueError(
                    "Cannot specify both split and fsplit. Please specify only one of the two.")
            else:
                self.splitvar = self.data[self.split]
                estimate_full_model = False
                estimate_split_model = True
                splitvar_name = self.split
        elif fsplit is not None:
            self.splitvar = self.data[fsplit]
            splitvar_name = fsplit
            estimate_split_model = True
        else:
            self.splitvar = None

        if self.splitvar is not None:
            self.split_categories = np.unique(self.splitvar)
            if splitvar_name not in self.data.columns:
                raise ValueError("Split variable " +
                                 self.splitvar + " not found in data.")
            if splitvar_name in self.var_dict.keys():
                raise ValueError("Split variable " + self.splitvar +
                                 " cannot be a fixed effect variable.")
            if self.splitvar.dtype.name != "category":
                self.splitvar = pd.Categorical(self.splitvar)

        if estimate_full_model:
            for _, fval in enumerate(fixef_keys):
                self.demeaned_data_dict[fval] = []
                self.dropped_data_dict[fval] = []
                self.yxz_name_dict[fval] = []
                data = self.data
                demeaned_data, dropped_data, yxz_name_dict= self._demean(
                    data, fval, ivars, drop_ref)
                self.demeaned_data_dict[fval].append(demeaned_data)
                self.dropped_data_dict[fval].append(dropped_data)
                self.yxz_name_dict[fval].append(yxz_name_dict)
        # here:
        if estimate_split_model:
            for _, fval in enumerate(fixef_keys):
                self.demeaned_data_dict[fval] = []
                self.dropped_data_dict[fval] = []
                self.yxz_name_dict[fval] = []
                for x in self.split_categories:
                    sub_data = self.data[x == self.splitvar]
                    demeaned_data, dropped_data, yxz_name_dict = self._demean(
                        sub_data, fval, ivars, drop_ref)
                    self.demeaned_data_dict[fval].append(demeaned_data)
                    self.dropped_data_dict[fval].append(dropped_data)
                    self.yxz_name_dict[fval].append(yxz_name_dict)

        # estimate models based on demeaned model matrix and dependent variables
        for _, fval in enumerate(self.fml_dict.keys()):
            model_splits = self.demeaned_data_dict[fval]
            for x, split in enumerate(model_splits):
                model_frames = model_splits[x]
                for _, fml in enumerate(model_frames):

                    # get the (demeaned) model frame. key is fml without fixed effects
                    model_frame = model_frames[fml]

                    # update formula with fixed effect. fval is "0" for no fixed effect
                    if fval == "0":
                        fml2 = fml
                    else:
                        fml2 = fml + "|" + fval

                    # formula log: add information on sample split
                    if self.splitvar is not None:
                        split_log = str(self.split_categories[x])
                        full_fml = fml2 + "| split =" + split_log
                    else:
                        split_log = None
                        full_fml = fml2

                    name_dict = self.yxz_name_dict[fval][0][fml]
                    depvar_name = name_dict["y_names"]
                    xvar_names = name_dict["x_names"]
                    if name_dict["z_names"] is None:
                        zvar_names = name_dict["x_names"]
                    else:
                        zvar_names = name_dict["z_names"]

                    Y = model_frame[depvar_name]
                    X = model_frame[xvar_names]
                    Z = model_frame[zvar_names]

                    colnames = X.columns
                    zcolnames = Z.columns

                    Y = Y.to_numpy()
                    X = X.to_numpy()
                    Z = Z.to_numpy()

                    N = X.shape[0]
                    k = X.shape[1]

                    # check for multicollinearity
                    if np.linalg.matrix_rank(X) < min(X.shape):
                        if self.ivars is not None:
                            raise ValueError("The design Matrix X does not have full rank for the regression with fml" + fml2 +
                                             ". The model is skipped. As you are running a regression via `i()` syntax, maybe you need to drop a level via i(var1, var2, ref = ...)?")
                        else:
                            raise ValueError(
                                "The design Matrizx X does not have full rank for the regression with fml" + fml2 + ". The model is skipped. ")

                    FEOLS = Feols(Y, X, Z)
                    FEOLS.is_iv = self.is_iv
                    FEOLS.fml = fml2
                    FEOLS.ssc_dict = self.ssc_dict
                    FEOLS.get_fit()
                    FEOLS.na_index = self.dropped_data_dict[fval][x][fml]
                    FEOLS.data = self.data.iloc[~self.data.index.isin(
                        FEOLS.na_index), :]
                    FEOLS.N = N
                    FEOLS.k = k
                    if fval != "0":
                        FEOLS.has_fixef = True
                        FEOLS.fixef = fval
                    else:
                        FEOLS.has_fixef = False

                    # FEOLS.get_nobs()

                    if vcov is None:
                        # iid if no fixed effects
                        if fval == "0":
                            vcov_type = "iid"
                        else:
                            # CRV1 inference, clustered by first fixed effect
                            first_fe = fval.split("+")[0]
                            vcov_type = {"CRV1": first_fe}
                    else:
                        vcov_type = vcov

                    FEOLS.vcov_log = vcov_type
                    FEOLS.split_log = x
                    FEOLS.get_vcov(vcov=vcov_type)
                    FEOLS.get_inference()
                    FEOLS.coefnames = colnames
                    if self.icovars is not None:
                        FEOLS.icovars = self.icovars
                    self.model_res[full_fml] = FEOLS

        return self

    def vcov(self, vcov: Union[str, Dict[str, str]]) -> None:
        '''
        Update regression inference "on the fly".
        By calling vcov() on a "Fixest" object, all inference procedures applied
        to the "Fixest" object are replaced with the variance covariance matrix specified via the method.
        Args:
            vcov: A string or dictionary specifying the type of variance-covariance matrix to use for inference.
                If a string, can be one of "iid", "hetero", "HC1", "HC2", "HC3".
                If a dictionary, it should have the format {"CRV1":"clustervar"} for CRV1 inference
                or {"CRV3":"clustervar"} for CRV3 inference.
        Returns:
            None
        '''

        self.vcov_log = vcov

        for model in list(self.model_res.keys()):

            fxst = self.model_res[model]
            fxst.vcov_log = vcov

            fxst.get_vcov(vcov=vcov)
            fxst.get_inference()

        return self

    def tidy(self, type: Optional[str] = None) -> Union[pd.DataFrame, str]:
        '''
        Returns the results of an estimation using `feols()` as a tidy Pandas DataFrame.
        Args:
            type : str, optional
                The type of output format to use. If set to "markdown", the resulting DataFrame
                will be returned in a markdown format with three decimal places. Default is None.
        Returns:
            pd.DataFrame or str
                A tidy DataFrame with the following columns:
                - fml: the formula used to generate the results
                - coefnames: the names of the coefficients
                - Estimate: the estimated coefficients
                - Std. Error: the standard errors of the estimated coefficients
                - t value: the t-values of the estimated coefficients
                - Pr(>|t|): the p-values of the estimated coefficients
                If `type` is set to "markdown", the resulting DataFrame will be returned as a
                markdown-formatted string with three decimal places.
        '''

        res = []
        for x in list(self.model_res.keys()):

            fxst = self.model_res[x]

            res.append(
                pd.DataFrame(
                    {
                        'fml': x,
                        'coefnames': fxst.coefnames,
                        'Estimate': fxst.beta_hat,
                        'Std. Error': fxst.se,
                        't value': fxst.tstat,
                        'Pr(>|t|)': fxst.pvalue
                    }
                )
            )

        res = pd.concat(res, axis=0).set_index('fml')
        if type == "markdown":
            return res.to_markdown(floatfmt=".3f")
        else:
            return res

    def summary(self) -> None:
        '''
        Prints a summary of the feols() estimation results for each estimated model.
        For each model, the method prints a header indicating the fixed-effects and the
        dependent variable, followed by a table of coefficient estimates with standard
        errors, t-values, and p-values.
        Returns:
            None
        '''

        for x in list(self.model_res.keys()):

            split = x.split("|")
            if len(split) > 1:
                fe = split[1]
            else:
                fe = None
            depvar = split[0].split("~")[0]
            fxst = self.model_res[x]
            df = pd.DataFrame(
                {
                    '': fxst.coefnames,
                    'Estimate': fxst.beta_hat,
                    'Std. Error': fxst.se,
                    't value': fxst.tstat,
                    'Pr(>|t|)': fxst.pvalue
                }
            )

            if fxst.is_iv:
                estimation_method = "IV"
            else:
                estimation_method = "OLS"


            print('###')
            print('')
            print('Model: ', estimation_method)
            print('Dep. var.: ', depvar)
            if fe is not None:
                print('Fixed effects: ', fe)
            # if fxst.split_log is not None:
            #    print('Split. var: ', self.split + ":" + fxst.split_log)
            print('Inference: ', fxst.vcov_log)
            print('Observations: ', fxst.N)
            print('')
            print(df.to_string(index=False))
            print('---')

    def coef(self) -> pd.DataFrame:
        '''
        Obtain the coefficients of the fitted models.
        Returns:
            A pd.DataFrame with coefficient names and Estimates. The key indicates which models the estimated statistic derives from.
        '''

        df = self.tidy()
        return df[['coefnames', 'Estimate']]

    def se(self)-> pd.DataFrame:
        '''
        Obtain the standard errors of the fitted models.

        Returns:
            A pd.DataFrame with coefficient names and standard error estimates. The key indicates which models the estimated statistic derives from.

        '''

        df = self.tidy()
        return df[['coefnames', 'Std. Error']]

    def tstat(self)-> pd.DataFrame:
        '''
        Obtain the t-statistics of the fitted models.

         Returns:
            A pd.DataFrame with coefficient names and estimated t-statistics. The key indicates which models the estimated statistic derives from.

        '''

        df = self.tidy()
        return df[['coefnames', 't value']]

    def pvalue(self) -> pd.DataFrame:
        '''
        Obtain the p-values of the fitted models.

        Returns:
            A pd.DataFrame with coefficient names and p-values. The key indicates which models the estimated statistic derives from.

        '''

        df = self.tidy()
        return df[['coefnames', 'Pr(>|t|)']]

    def iplot(self, alpha: float = 0.05, figsize: tuple = (10, 10), yintercept: Union[int, str, None] = None, xintercept: Union[int, str, None] = None, rotate_xticks: int = 0) -> None:
        '''
        Plot model coefficients with confidence intervals for variable interactions specified via the `i()` syntax.
        Args:
            alpha: float, optional. The significance level for the confidence intervals. Default is 0.05.
            figsize: tuple, optional. The size of the figure. Default is (10, 10).
            yintercept: int or str (for a categorical x axis). The value at which to draw a horizontal line.
            xintercept: int or str (for a categorical y axis). The value at which to draw a vertical line.
        Returns:
            None
        '''

        ivars = self.icovars

        if ivars is None:
            raise ValueError(
                "The estimated models did not have ivars / 'i()' model syntax. In consequence, the '.iplot()' method is not supported.")

        ivars_keys = self.ivars.keys()
        if ivars_keys is not None:
            ref = list(ivars_keys)[0]
        else:
            ref = None

        if "Intercept" in ivars:
            ivars.remove("Intercept")

        df = self.tidy()

        df = df[df.coefnames.isin(ivars)]
        models = df.index.unique()

        _coefplot(
            models=models,
            figsize=figsize,
            alpha=alpha,
            yintercept=yintercept,
            xintercept=xintercept,
            df=df,
            is_iplot=True
        )

    def coefplot(self, alpha: float = 0.05, figsize: tuple = (5, 2), yintercept: int = 0, figtitle: str = None, figtext: str = None, rotate_xticks: int = 0) -> None:
        '''
        Plot estimation results. The plot() method is only defined for single regressions.
        Args:
            alpha (float): the significance level for the confidence intervals. Default is 0.05.
            figsize (tuple): the size of the figure. Default is (5, 2).
            yintercept (float): the value of the y-intercept. Default is 0.
            figtitle (str): the title of the figure. Default is None.
            figtext (str): the text at the bottom of the figure. Default is None.
        Returns:
            None
        '''

        df = self.tidy()
        models = df.index.unique()

        _coefplot(
            models=models,
            figsize=figsize,
            alpha=alpha,
            yintercept=yintercept,
            xintercept=None,
            df=df,
            is_iplot=False,
            rotate_xticks=rotate_xticks
        )

    def wildboottest(self, B, param: Union[str, None] = None, weights_type: str = 'rademacher', impose_null: bool = True, bootstrap_type: str = '11', seed: Union[str, None] = None, adj: bool = True, cluster_adj: bool = True) -> pd.DataFrame:

        '''
        Run a wild cluster bootstrap for all regressions in the Fixest object.

        Args:

            B (int): The number of bootstrap iterations to run
            param (Union[str, None], optional): A string of length one, containing the test parameter of interest. Defaults to None.
            weights_type (str, optional): The type of bootstrap weights. Either 'rademacher', 'mammen', 'webb' or 'normal'.
                                'rademacher' by default. Defaults to 'rademacher'.
            impose_null (bool, optional): Should the null hypothesis be imposed on the bootstrap dgp, or not?
                                Defaults to True.
            bootstrap_type (str, optional):A string of length one. Allows to choose the bootstrap type
                                to be run. Either '11', '31', '13' or '33'. '11' by default. Defaults to '11'.
            seed (Union[str, None], optional): Option to provide a random seed. Defaults to None.

        Returns:
            A pd.DataFrame with bootstrapped t-statistic and p-value. The index indicates which model the estimated statistic derives from.
        '''


        res = []
        for x in list(self.model_res.keys()):

            fxst = self.model_res[x]

            if hasattr(fxst, 'clustervar'):
                cluster = fxst.clustervar
            else:
                cluster = None

            boot_res = fxst.get_wildboottest(B, cluster, param,  weights_type, impose_null, bootstrap_type, seed, adj, cluster_adj)

            pvalue = boot_res["pvalue"]
            tstat = boot_res["statistic"]


            res.append(
                pd.Series(
                    {
                        'fml': x,
                        'param':param,
                        't value': tstat,
                        'Pr(>|t|)': pvalue
                    }
                )
            )

        res = pd.concat(res, axis=1).T.set_index('fml')

        return res


def _coefplot(models: List, df: pd.DataFrame, figsize: Tuple[int, int], alpha: float, yintercept: Optional[int] = None,
              xintercept: Optional[int] = None, is_iplot: bool = False,
              rotate_xticks: float = 0) -> None:
    """
        Plot model coefficients with confidence intervals.
        Args:
            models (list): A list of fitted models.
            figsize (tuple): The size of the figure.
            alpha (float): The significance level for the confidence intervals.
            yintercept (int or None): The value at which to draw a horizontal line on the plot.
            xintercept (int or None): The value at which to draw a vertical line on the plot.
            df (pandas.DataFrame): The dataframe containing the data used for the model fitting.
            is_iplot (bool): If True, plot variable interactions specified via the `i()` syntax.
            rotate_xticks (float): The angle in degrees to rotate the xticks labels. Default is 0 (no rotation).
        Returns:
        None
    """

    if len(models) > 1:

        fig, ax = plt.subplots(len(models), gridspec_kw={
                               'hspace': 0.5}, figsize=figsize)

        for x, model in enumerate(models):

            df_model = df.xs(model)
            coef = df_model["Estimate"].values
            conf_l = coef - \
                df_model["Std. Error"].values * norm.ppf(1 - alpha / 2)
            conf_u = coef + \
                df_model["Std. Error"].values * norm.ppf(1 - alpha / 2)
            coefnames = df_model["coefnames"].values.tolist()

            # could be moved out of the for loop, as the same ivars for all
            # models.

            if is_iplot == True:
                fig.suptitle("iplot")
                coefnames = [(i) for string in coefnames for i in re.findall(
                    r'\[T\.([\d\.\-]+)\]', string)]

            # in the future: add reference category
            # if ref is not None:
            #    coef = np.append(coef, 0)
            #    conf_l = np.append(conf_l, 0)
            #    conf_u = np.append(conf_u, 0)
            #    coefnames = np.append(coefnames, ref)

            ax[x].scatter(coefnames, coef, color="b", alpha=0.8)
            ax[x].scatter(coefnames, conf_u, color="b",
                          alpha=0.8, marker="_", s=100)
            ax[x].scatter(coefnames, conf_l, color="b",
                          alpha=0.8, marker="_", s=100)
            ax[x].vlines(coefnames, ymin=conf_l,
                         ymax=conf_u, color="b", alpha=0.8)
            if yintercept is not None:
                ax[x].axhline(yintercept, color='red',
                              linestyle='--', alpha=0.5)
            if xintercept is not None:
                ax[x].axvline(xintercept, color='red',
                              linestyle='--', alpha=0.5)
            ax[x].set_ylabel('Coefficients')
            ax[x].set_title(model)
            ax[x].tick_params(axis='x', rotation=rotate_xticks)

    else:

        fig, ax = plt.subplots(figsize=figsize)

        model = models[0]

        df_model = df.xs(model)
        coef = df_model["Estimate"].values
        conf_l = coef - df_model["Std. Error"].values * norm.ppf(1 - alpha / 2)
        conf_u = coef + df_model["Std. Error"].values * norm.ppf(1 - alpha / 2)
        coefnames = df_model["coefnames"].values.tolist()

        if is_iplot == True:
            fig.suptitle("iplot")
            coefnames = [(i) for string in coefnames for i in re.findall(
                r'\[T\.([\d\.\-]+)\]', string)]

        # in the future: add reference category
        # if ref is not None:
        #    coef = np.append(coef, 0)
        #    conf_l = np.append(conf_l, 0)
        #    conf_u = np.append(conf_u, 0)
        #    coefnames = np.append(coefnames, ref)

            # c = next(color)

        ax.scatter(coefnames, coef, color="b", alpha=0.8)
        ax.scatter(coefnames, conf_u, color="b", alpha=0.8, marker="_", s=100)
        ax.scatter(coefnames, conf_l, color="b", alpha=0.8, marker="_", s=100)
        ax.vlines(coefnames, ymin=conf_l, ymax=conf_u, color="b", alpha=0.8)
        if yintercept is not None:
            ax.axhline(yintercept, color='red', linestyle='--', alpha=0.5)
        if xintercept is not None:
            ax.axvline(xintercept, color='red', linestyle='--', alpha=0.5)
        ax.set_ylabel('Coefficients')
        ax.set_title(model)
        ax.tick_params(axis='x', rotation=rotate_xticks)

        plt.show()
        plt.close()



import numpy as np

def ssc(adj=True, fixef_k="none", cluster_adj=True, cluster_df="conventional"):
    '''
    Set the small sample correction factor applied in `get_ssc()`
    Parameters:
        adj: bool, default True
            If True, applies a small sample correction of (N-1) / (N-k) where N is the number of observations
            and k is the number of estimated coefficients excluding any fixed effects projected out in either fixest::feols() or lfe::felm().
        fixef_k: str, default "none"
            Equal to 'none': the fixed effects parameters are discarded when calculating k in (N-1) / (N-k).
        cluster_adj: bool, default True
            If True, a cluster correction G/(G-1) is performed, with G the number of clusters.
        cluster_df: str, default "conventional"
            Controls how "G" is computed for multiway clustering if cluster_adj = True.
            Note that the covariance matrix in the multiway clustering case is of
            the form V = V_1 + V_2 - V_12. If "conventional", then each summand G_i
            is multiplied with a small sample adjustment G_i / (G_i - 1). If "min",
            all summands are multiplied with the same value, min(G) / (min(G) - 1)
    Returns:
        A dictionary with encoded info on how to form small sample corrections
    '''

    if adj not in [True, False]:
        raise ValueError("adj must be True or False.")
    if fixef_k not in ["none"]:
        raise ValueError("fixef_k must be 'none'.")
    if cluster_adj not in [True, False]:
        raise ValueError("cluster_adj must be True or False.")
    if cluster_df not in ["conventional", "min"]:
        raise ValueError("cluster_df must be 'conventional' or 'min'.")

    res = {'adj': adj, 'fixef_k': fixef_k,
           'cluster_adj': cluster_adj, 'cluster_df': cluster_df}

    return res


def get_ssc(ssc_dict, N, k, G, vcov_sign, vcov_type):
    """
    Compute small sample adjustment factors

    Args:
    - ssc_dict: An dictionariy created via the ssc() function
    - N: The number of observations
    - k: The number of estimated parameters
    - G: The number of clusters
    - vcov_sign: A vector that helps create the covariance matrix
    - vcov_type: Either "iid", "hetero" or "CRV"

    Returns:
    - A small sample adjustment factor
    """

    adj = ssc_dict['adj']
    fixef_k = ssc_dict['fixef_k']
    cluster_adj = ssc_dict['cluster_adj']
    cluster_df = ssc_dict['cluster_df']

    cluster_adj_value = 1
    adj_value = 1

    if vcov_type == "hetero":
        if adj:
            #adj_value = (N - 1) / (N - k)
            adj_value = N / (N-k)
        else:
            #adj_value = 1
            adj_value = N / (N-1)
    elif vcov_type in ["iid", "CRV"]:
        if adj:
            adj_value = (N - 1) / (N - k)
        else:
            adj_value = 1
    else:
        raise ValueError("vcov_type must be either iid, hetero or CRV.")

    if vcov_type == "CRV":

        if cluster_adj:
            if cluster_df == 'conventional':
                cluster_adj_value = G / (G - 1)
            elif cluster_df == "min":
                G = np.min(G)
                cluster_adj_value = G / (G - 1)
            else:
                raise ValueError("cluster_df is neither conventional nor min.")

    return adj_value * cluster_adj_value * vcov_sign



import numpy as np
import pandas as pd

def get_data(seed = 1234):

    '''
    create a random example data set
    '''

    # create data

    np.random.seed(seed)

    N = 2000
    k = 20
    G = 25
    X = np.random.normal(0, 1, N * k).reshape((N,k))
    X = pd.DataFrame(X)
    X[1] = np.random.choice(list(range(0, 5)), N, True)
    X[2] = np.random.choice(list(range(0, 10)), N, True)
    X[3] = np.random.choice(list(range(0, 10)), N, True)
    X[4] = np.random.normal(0, 1, N)

    beta = np.random.normal(0,1,k)
    beta[0] = 0.005
    u = np.random.normal(0,1,N)
    Y = 1 + X @ beta + u
    cluster = np.random.choice(list(range(0,G)), N)

    Y = pd.DataFrame(Y)
    Y.rename(columns = {0:'Y'}, inplace = True)
    X = pd.DataFrame(X)

    data = pd.concat([Y, X], axis = 1)
    data.rename(columns = {0:'X1', 1:'X2', 2:'X3', 3:'X4', 4:'X5'}, inplace = True)

    data['group_id'] = cluster
    data['Y2'] = data.Y + np.random.normal(0, 1, N)

    data['Y'][0] = np.nan
    data['X1'][1] = np.nan

    data["Z1"] = data["X1"] + np.random.normal(0, 1, data.shape[0])
    #data['X2'][2] = np.nan


    return data



import re

class FixedEffectInteractionError(Exception):
    pass

class CovariateInteractionError(Exception):
    pass

class DuplicateKeyError(Exception):
    pass


class FixestFormulaParser:


    """
    A class for parsing a formula string into its individual components.

    Attributes:
        depvars (list): A list of dependent variables in the formula.
        covars (list): A list of covariates in the formula.
        fevars (list): A list of fixed effect variables in the formula.
        covars_fml (str): A string representation of the covariates in the formula.
        fevars_fml (str): A string representation of the fixed effect variables in the formula.

    Methods:
        __init__(self, fml): Constructor method that initializes the object with a given formula.
        get_fml_dict(self): Returns a dictionary of all fevars & formula without fevars.
        get_var_dict(self): Returns a dictionary of all fevars and list of covars and depvars used in regression with those fevars.

    """

    def __init__(self, fml):

        """
        Constructor method that initializes the object with a given formula.

        Args:
        fml (str): A two-formula string in the form "Y1 + Y2 ~ X1 + X2 | FE1 + FE2".

        Returns:
            None
        """

        #fml =' Y + Y2 ~  i(X1, X2) |csw0(X3, X4)'

        # Clean up the formula string
        fml = "".join(fml.split())

        # Split the formula string into its components
        fml_split = fml.split('|')
        depvars, covars = fml_split[0].split("~")

        if len(fml_split) == 1:
            fevars = "0"
            endogvars = None
            instruments = None
        elif len(fml_split) == 2:
            if "~" in fml_split[1]:
                fevars = "0"
                endogvars, instruments = fml_split[1].split("~")
            else:
                fevars = fml_split[1]
                endogvars = None
                instruments = None
        elif len(fml_split) == 3:
            fevars = fml_split[1]
            endogvars, instruments = fml_split[2].split("~")

        if endogvars is not None:
            if len(endogvars) > len(instruments):
                raise ValueError("The IV system is underdetermined. Only fully determined systems are allowed. Please provide as many instruments as endogenous variables.")
            elif len(endogvars) < len(instruments):
                raise ValueError("The IV system is overdetermined. Only fully determined systems are allowed. Please provide as many instruments as endogenous variables.")
            else:
                pass

        # Parse all individual formula components into lists
        self.depvars = depvars.split("+")
        self.covars = _unpack_fml(covars)
        self.fevars = _unpack_fml(fevars)
        # no fancy syntax for endogvars, instruments allowed
        self.endogvars = endogvars
        self.instruments = instruments

        if instruments is not None:
            self.is_iv = True
            # all rhs variables for the first stage (endog variable replaced with instrument)
            first_stage_covars_list = covars.split("+")
            first_stage_covars_list[first_stage_covars_list.index(endogvars)] = instruments
            self.first_stage_covars_list = "+".join(first_stage_covars_list)
            self.covars_first_stage = _unpack_fml(self.first_stage_covars_list)
            self.depvars_first_stage = endogvars
        else:
            self.is_iv = False
            self.covars_first_stage = None
            self.depvars_first_stage = None

        if self.covars.get("i") is not None:
            self.ivars = dict()
            i_split = self.covars.get("i")[-1].split("=")
            if len(i_split) > 1:
                ref = self.covars.get("i")[-1].split("=")[1]
                ivar_list = self.covars.get("i")[:-1]
                self.covars["i"] = self.covars.get("i")[:-1]
            else:
                ref = None
                ivar_list = self.covars.get("i")

            self.ivars[ref] = ivar_list

        else:
            self.ivars = None


        # Pack the formula components back into strings
        self.covars_fml = _pack_to_fml(self.covars)
        self.fevars_fml = _pack_to_fml(self.fevars)
        if instruments is not None: 
            self.covars_first_stage_fml = _pack_to_fml(self.covars_first_stage)
        else: 
            self.covars_first_stage_fml = None
        #if "^" in self.covars:
        #    raise CovariateInteractionError("Please use 'i()' or ':' syntax to interact covariates.")

        #for  x in ["i", ":"]:
        #    if x in self.fevars:
        #        raise FixedEffectInteractionError("Interacting fixed effects via", x, " is not allowed. Please use '^' to interact fixed effects.")





    def get_fml_dict(self, iv = False):

        """
        Returns a dictionary of all fevars & formula without fevars. The keys are the fixed effect variable combinations.
        The values are lists of formula strings that do not include the fixed effect variables.

        Args:
            iv (bool): If True, the formula dictionary will be returned for the first stage of an IV regression.
                       If False, the formula dictionary will be returned for the second stage of an IV regression / OLS regression.
        Returns:
            dict: A dictionary of the form {"fe1+fe2": ['Y1 ~ X', 'Y2~X'], "fe1+fe3": ['Y1 ~ X', 'Y2~X']} where
            the keys are the fixed effect variable combinations and the values are lists of formula strings
            that do not include the fixed effect variables.
            If IV is True, creates an instance named fml_dict_iv. Otherwise, creates an instance named fml_dict.
        """


        fml_dict = dict()
        for fevar in self.fevars_fml:
            res = []
            for depvar in self.depvars:
                if iv:
                    for covar in self.covars_first_stage_fml:
                        res.append(depvar + '~' + covar)
                else:
                    for covar in self.covars_fml:
                        res.append(depvar + '~' + covar)
            fml_dict[fevar] = res

        if iv:
            self.fml_dict_iv = fml_dict
        else:
            self.fml_dict = fml_dict

    def _transform_fml_dict(self, iv = False):

        fml_dict2 = dict()

        if iv:

            for fe in self.fml_dict_iv.keys():

                fml_dict2[fe] = dict()

                for fml in self.fml_dict_iv.get(fe):
                    depvars, covars = fml.split("~")
                    if fml_dict2[fe].get(depvars) is None:
                        fml_dict2[fe][depvars] = [covars]
                    else:
                        fml_dict2[fe][depvars].append(covars)
        else:

          for fe in self.fml_dict.keys():

              fml_dict2[fe] = dict()

              for fml in self.fml_dict.get(fe):
                  depvars, covars = fml.split("~")
                  if fml_dict2[fe].get(depvars) is None:
                      fml_dict2[fe][depvars] = [covars]
                  else:
                      fml_dict2[fe][depvars].append(covars)

        if iv:
            self.fml_dict2_iv = fml_dict2
        else:
            self.fml_dict2 = fml_dict2



    def get_var_dict(self, iv = False):

        """
        Create a dictionary of all fevars and list of covars and depvars used in regression with those fevars.
        The keys are the fixed effect variable combinations. The values are lists of variables (dependent variables and covariates) of
        the resespective regressions.

        Args:
            iv (bool): If True, the formula dictionary will be returned for the first stage of an IV regression.

        Returns:
            dict: A dictionary of the form {"fe1+fe2": ['Y1', 'X1', 'X2'], "fe1+fe3": ['Y1', 'X1', 'X2']} where
            the keys are the fixed effect variable combinations and the values are lists of variables
            (dependent variables and covariates) used in the regression with those fixed effect variables.

        """
        var_dict = dict()
        if iv:
            for fevar in self.fevars_fml:
                var_dict[fevar] = _flatten_list(self.depvars) + _flatten_list(list(self.covars_first_stage.values()))

        else:
            for fevar in self.fevars_fml:
                var_dict[fevar] = _flatten_list(self.depvars) + _flatten_list(list(self.covars.values()))

        if iv:
            self.var_dict_iv = var_dict
        else:
            self.var_dict = var_dict


def _unpack_fml(x):

    '''
    Given a formula string `x` - e.g. 'X1 + csw(X2, X3)' - , splits it into its constituent variables and their types (if any),
    and returns a dictionary containing the result. The dictionary has the following keys: 'constant', 'sw', 'sw0', 'csw'.
    The values are lists of variables of the respective type.

    Parameters:
    -----------
    x : str
        The formula string to unpack.

    Returns:
    --------
    res_s : dict
        A dictionary containing the unpacked formula. The dictionary has the following keys:
            - 'constant' : list of str
                The list of constant (i.e., non-switched) variables in the formula.
            - 'sw' : list of str
                The list of variables that have a regular switch (i.e., 'sw(var1, var2, ...)' notation) in the formula.
            - 'sw0' : list of str
                The list of variables that have a 'sw0(var1, var2, ..)' switch in the formula.
            - 'csw' : list of str or list of lists of str
                The list of variables that have a 'csw(var1, var2, ..)' switch in the formula.
                Each element in the list can be either a single variable string, or a list of variable strings
                if multiple variables are listed in the switch.
            - 'csw0' : list of str or list of lists of str
                The list of variables that have a 'csw0(var1,var2,...)' switch in the formula.
                Each element in the list can be either a single variable string, or a list of variable strings
                if multiple variables are listed in the switch.

    Raises:
    -------
    ValueError:
        If the switch type is not one of 'sw', 'sw0', 'csw', or 'csw0'.

    Example:
    --------
    >>> _unpack_fml('a+sw(b)+csw(x1,x2)+sw0(d)+csw0(y1,y2,y3)')
    {'constant': ['a'],
     'sw': ['b'],
     'csw': [['x1', 'x2']],
     'sw0': ['d'],
     'csw0': [['y1', 'y2', 'y3']]}
    '''


    # Split the formula into its constituent variables
    var_split = x.split("+")

    res_s = dict()
    res_s['constant'] = []

    for var in var_split:

        # Check if this variable contains a switch
        varlist, sw_type = _find_sw(var)

        # If there's no switch, just add the variable to the list
        if sw_type is None:
            res_s['constant'].append(var)

        # If there's a switch, unpack it and add it to the list
        else:
            if sw_type in ['sw', 'sw0', 'csw', 'csw0', 'i']:
                _check_duplicate_key(res_s, sw_type)
                res_s[sw_type] = varlist
            else:
                raise ValueError("Unsupported switch type")

    # Sort the list by type (strings first, then lists)
    #res_s.sort(key=lambda x: 0 if isinstance(x, str) else 1)

    return res_s




def _pack_to_fml(unpacked):
    """
    Given a dictionary of "unpacked" formula variables, returns a string containing formulas. An "unpacked" formula is a
    deparsed formula that allows for multiple estimations.

    Parameters
    ----------
    unpacked : dict
        A dictionary of unpacked formula variables. The dictionary has the following keys:
            - 'constant' : list of str
                The list of constant (i.e., non-switched) variables in the formula.
            - 'sw' : list of str
                The list of variables that have a regular switch (i.e., 'sw(var1, var2, ...)' notation) in the formula.
            - 'sw0' : list of str
                The list of variables that have a 'sw0(var1, var2, ..)' switch in the formula.
            - 'csw' : list of str or list of lists of str
                The list of variables that have a 'csw(var1, var2, ..)' switch in the formula.
                Each element in the list can be either a single variable string, or a list of variable strings
                if multiple variables are listed in the switch.
            - 'csw0' : list of str or list of lists of str
    """

    res = dict()

    # add up all constant variables
    if 'constant' in unpacked:
         res['constant'] = unpacked['constant']
    else:
        res['constant'] = []

    if 'i' in unpacked:
       if res['constant']:
           res['constant'] =  res['constant'] + [":".join(unpacked['i'])]
       else:
           res['constant'] = [":".join(unpacked['i'])]

    # add up all variable constants (only required for csw)
    if "csw" in unpacked:
        res['variable'] = unpacked['csw']
        variable_type = "csw"
    elif "csw0" in unpacked:
        res['variable'] = unpacked['csw0']
        variable_type = "csw0"
    elif "sw" in unpacked:
        res['variable'] = unpacked['sw']
        variable_type = "sw"
    elif "sw0" in unpacked:
        res['variable'] = unpacked['sw0']
        variable_type = "sw0"
    else:
        res['variable'] = []
        variable_type = None

    if res['constant']:
        const_fml = "+".join(res['constant'])
    else:
        const_fml = []

    variable_fml = []
    if res['variable']:
        if variable_type in ['csw', 'csw0']:
            variable_fml = [ "+".join(res['variable'][:i+1]) for i in range(len(res['variable']))]
        else:
            variable_fml = [res['variable'][i] for i in range(len(res['variable']))]
        if variable_type in ['sw0', 'csw0']:
            variable_fml = ['0'] + variable_fml


    fml_list = []
    if variable_fml:
        if const_fml:
            fml_list = [ const_fml + "+" + variable_fml[i] for i in range(len(variable_fml)) if variable_fml[i] != "0"]
            if variable_type in ['sw0', 'csw0']:
                fml_list = [const_fml] + fml_list
        else:
            fml_list = variable_fml
    else:
        if const_fml:
            fml_list = const_fml
        else:
            raise Exception("Not a valid formula provided.")

    if not isinstance(fml_list, list):
        fml_list = [fml_list]

    return fml_list



def _find_sw(x):
    """
    Search for matches in a string. Matches are either 'sw', 'sw0', 'csw', 'csw0', or 'i'. If a match is found, returns a
    tuple containing a list of the elements found and the type of match. Otherwise, returns the original string and None.

    Args:
        x (str): The string to search for matches in.

    Returns:
        (list[str] or str, str or None): If any matches were found, returns a tuple containing
        a list of the elements found and the type of match (either 'sw', 'sw0', 'csw', or 'csw0').
        Otherwise, returns the original string and None.

    Example:
        _find_sw('sw(var1, var2)') -> (['var1', ' var2'], 'sw')
    """

    # Search for matches in the string
    sw_match = re.findall(r"sw\((.*?)\)", x)
    csw_match = re.findall(r"csw\((.*?)\)", x)
    sw0_match = re.findall(r"sw0\((.*?)\)", x)
    csw0_match = re.findall(r"csw0\((.*?)\)", x)
    i_match = re.findall(r"i\((.*?)\)", x)


    # Check for sw matches
    if sw_match:
        if csw_match:
            return csw_match[0].split(","), "csw"
        else:
            return sw_match[0].split(","), "sw"

    # Check for sw0 matches
    elif sw0_match:
        if csw0_match:
            return csw0_match[0].split(","), "csw0"
        else:
            return sw0_match[0].split(","), "sw0"

    elif i_match:
        return i_match[0].split(","), "i"

    # No matches found
    else:
        return x, None


def _flatten_list(lst):

    """
    Flattens a list that may contain sublists.

    Args:
        lst (list): A list that may contain sublists.

    Returns:
        list: A flattened list with no sublists.

    Examples:
        >>> flatten_list([[1, 2, 3], 4, 5])
        [1, 2, 3, 4, 5]
        >>> flatten_list([1, 2, 3])
        [1, 2, 3]
    """

    flattened_list = []
    for i in lst:
        if isinstance(i, list):
            flattened_list.extend(_flatten_list(i))
        else:
            flattened_list.append(i)
    return flattened_list


def _check_duplicate_key(my_dict, key):

    '''
    Checks if a key already exists in a dictionary. If it does, raises a DuplicateKeyError. Otherwise, does nothing.

    Args:
        my_dict (dict): The dictionary to check for duplicate keys.
        key (str): The key to check for in the dictionary.

    Returns:
        None
    '''

    if key == 'i' and 'i' in my_dict:
        raise DuplicateKeyError("Duplicate key found: " + key + ". Fixed effect syntax i() can only be used once in the input formula.")
    else:
        for key in ['sw', 'csw', 'sw0', 'csw0']:
            if key in my_dict:
                raise DuplicateKeyError("Duplicate key found: " + key + ". Multiple estimation syntax can only be used once on the rhs of the two-sided formula.")
            else:
                None
